#include <string>
#include <iostream>
#include "neocca.hh"
#include "neoports.hh"
#include "AllEGPorts.hh"
#include "SimpleStamper.hh"
		   
//
// some things need to be adjusted for the
// latest sidl Component destructor callback
// services function.

using namespace std;
using namespace eg::neo::ports;
using namespace neo::cca;

eg::neo::SimpleStamper::SimpleStamper()
{
  svc = 0;
  time = 0;
}

eg::neo::SimpleStamper::~SimpleStamper()
{
  svc = 0;
  time = 0;
}

void ::eg::neo::SimpleStamper::releaseServices(Services* cc)
throw (Exception)
{
	if (cc == 0 || cc != svc) {
		throw Exception( Nonstandard, "::eg::neo::SimpleStamper::releaseServices: bogus input");
	}
	//cerr << "SimpleStamper::releaseServices called." << endl;
	// time to shut down
	if (time != 0) {
		svc->releasePort("time");
		time = 0;
	}
	svc->unregisterUsesPort("time");
	svc->unregisterUsesPort("out_port");
	svc->removeProvidesPort("in_port");
	svc->removeProvidesPort("go");
	svc = 0;
	return;
}

void ::eg::neo::SimpleStamper::setServices(Services *cc)
throw (Exception)
{

  // two boiler-plate defensive programming cases follow.
  // component implementations which omit them will run
  // safely, but errors which could have been
  // detected here appear may more cryptically later.
	// for simplicity replace with 
	// assert(cc!=0&&svc==0);
  
  if (svc != 0) {
    // we're running already and someone tried to start us again.
    // or some bozo forgot to init svc to 0 in our constructor.
    throw Exception(
      "SimpleStamper: someone tried to setServices twice. violates cca spec.");
    return;
  }
  if (cc == 0 ) {
    // framework is confused.
    // we're not running yet.
    // we're running already and someone tried to start us.
    throw Exception(
      "SimpleStamper: setServices(null) called. violates cca spec.");
    return; 
  }

  if ( svc == 0 ) {
    // it's time to start up.
  
    svc = cc;
    svc->addProvidesPort(this, "in_port", "eg.neo.StringOutPort");
    svc->addProvidesPort(this, "go", "neo.cca.ports.GoPort");
    svc->registerUsesPort("time", "eg.neo.TimeSourcePort");
    svc->registerUsesPort("out_port", "eg.neo.StringOutPort");
    svc->registerForRelease(this);
    return;
  }

}

int 
::eg::neo::SimpleStamper::go(){
  Port * p = svc->getPort("out_port");
  StringOutPort *otherSOP = 0;
  if (p==0) {
    cerr << "SimpleStamper::go: out_port unavailable." << endl;
    cerr << "SimpleStamper::go: shouldn't we experience an exception here?" << endl;
  } else {
    otherSOP = dynamic_cast < StringOutPort * >(p);
  }
  if ( otherSOP == 0 ) {
    cerr << "SimpleStamper::go: out_port unavailable (failed cast)." << endl;
    return 0;
  }
  otherSOP->setString("SimpleStamper::go(): called non-self setString.");
  svc->releasePort("out_port");
  return 0;
}

void 
::eg::neo::SimpleStamper::setString(const std::string & s)
{

  Port *p = 0;
  try {
    p = svc->getPort("time");
  } catch ( ::neo::cca::Exception ne) {
	  cerr << "SimpleStamper::setString: caught exception from getport(time) " <<
		  ne.what() << endl;
	  p = 0;
  }
  if (p==0) {
    cerr << "SimpleStamper::setString: time Port unavailable." << endl;
  } else {
    time = dynamic_cast < TimeSourcePort * >(p);
    if ( time == 0 ) {
      cerr << "SimpleStamper::setString: time Port unavailable (failed cast)." << endl;
    }
  }
  if ( time == 0 ) {
    cout << "SimpleStamper: " << s << "@unknown_time" << endl;
    return;
  }

  string t = time->getTime();
  string str = s + "@" + t;
  cout << "SimpleStamper: " << str << endl;

  if (p != 0 ) {
    svc->releasePort("time");
  }
  time = 0;
}

//////////////////////////// wrapper 
extern "C" {

neo::cca::Component *create_eg_neo_SimpleStamper() {
        neo::cca::Component *comp_ptr;
        eg::neo::SimpleStamper *component;
        component = new eg::neo::SimpleStamper();
	assert(component != 0);
        comp_ptr = dynamic_cast<neo::cca::Component *>(component);
	assert(comp_ptr != 0);
        return comp_ptr;
}


}

