#ifdef __GNUG__
#if (__GNUG__==2)
#include <sstream>
#endif
#endif
#include <iostream>
#include <map>
#include <neocca.hh>
#include <neoports.hh>
#include <neosupport.hh>

/** This function handles the configuring of parameters when all we know is the
   info available from a ccaffeine script (which hides parameter and port types).
   If component port and config information were ubiquitous at code generation time, this
   wouldn't be needed.
	@param compName name of the component being configured.
	@param port name of the port being configured.
	@param comp ComponentID of the component being configured.
	@param bs BuilderService controlling the frame.
	@param services Services of the component/driver doing the parameter setting.
	@param values  map of keys and values to assign.
*/
void 
neo::support::helpers::setParameters(const std::string compName, const std::string port, neo::cca::ComponentID_shared comp, neo::cca::ports::BuilderService *bs, neo::cca::Services *services, std::map< std::string, std::string > & values)
{
	neo::cca::ComponentID_shared myself = services->getComponentID();
	neo::cca::TypeMap_shared dummy;
	neo::cca::TypeMap_shared pProps = bs->getPortProperties(comp, port);
	std::string portType = pProps->getString("cca.portType", "undefined");
 	neo::cca::Port *p_param = 0;
	std::string pname = compName;
	pname += "_";
	pname += port;
	if ( portType == "neo::cca::ports::BasicParameterPort" ||
		portType == "::neo::cca::ports::BasicParameterPort" ||
		portType ==  "neo.cca.ports.BasicParameterPort") 
	{
		neo::cca::ports::BasicParameterPort * bpp = 0;
		services->registerUsesPort(pname, "neo::cca::ports::BasicParameterPort", dummy);
		neo::cca::ConnectionID_shared bppconn =
			bs->connect(myself, pname, comp, port);
		p_param = services->getPort(pname);
		bpp = dynamic_cast< neo::cca::ports::BasicParameterPort *>(p_param);
		if (bpp != 0) {
			neo::cca::TypeMap_shared params = bpp->readConfigurationMap();
			neo::support::helpers::setParamsFromString(params, values);
			bpp->writeConfigurationMap(params);
		} else {
			std::string msg = "Port ";
			msg += compName; msg += "."; msg += port;
			msg += " is not of expected type BasicParameterPort (cast fail)";
			throw neo::cca::Exception(neo::cca::BadPortType, msg);
		}
		services->releasePort(pname);
		bs->disconnect(bppconn, 0.0);
		services->unregisterUsesPort(pname);
		return;
	}
	if ( portType == "neo::cca::ports::ParameterPort" ||
		portType == "::neo::cca::ports::ParameterPort" ||
		portType ==  "neo.cca.ports.ParameterPort") 
	{
		neo::cca::ports::ParameterPort * pp = 0;
		services->registerUsesPort(pname, "neo::cca::ports::ParameterPort", dummy);
		neo::cca::ConnectionID_shared ppconn =
			bs->connect(myself, pname, comp, port);
		p_param = services->getPort(pname);
		pp = dynamic_cast< neo::cca::ports::ParameterPort *>(p_param);
		if (pp != 0) {
			neo::cca::TypeMap_shared params = pp->readConfigurationMap();
			neo::support::helpers::setParamsFromString(params, values);
			pp->writeConfigurationMap(params);
		} else {
			std::string msg = "Port ";
			msg += compName; msg += "."; msg += port;
			msg += " is not of expected type ParameterPort (cast fail)";
			throw neo::cca::Exception(neo::cca::BadPortType, msg);
		}
		services->releasePort(pname);
		bs->disconnect(ppconn, 0.0);
		services->unregisterUsesPort(pname);
		return;
	}
	std::string msg = "Port ";
	msg += compName; msg += "."; msg += port;
	msg += " is not a parameter port. seems to be a ";
	msg += portType;
	throw neo::cca::Exception(neo::cca::BadPortType, msg);
}

/** For keys in both tm and values, copy values to tm in the
appropriate format.
*/
void
neo::support::helpers::setParamsFromString(neo::cca::TypeMap_shared tm, std::map< std::string, std::string > & values)
{
	std::map< std::string, std::string >::const_iterator it;
	std::string key;
	std::string val;
	neo::cca::Type dt;
	for ( it = values.begin(); it != values.end(); ++it ){
            	key =  it->first;
		val = it->second;
		dt = tm->typeOf(key);
		if (dt != neo::cca::NoType) {
			transformString(key, dt, val, tm);
		}
	}
}

int
neo::support::helpers::invokeGo( const std::string comp, const std::string port, neo::cca::ComponentID_shared target, neo::cca::Services *services, neo::cca::ports::BuilderService *bs)
{
	std::string pname = comp;
	pname += "_";
	pname += port;
	neo::cca::Port * p_go;
	services->registerUsesPort(pname, "neo::cca::ports::GoPort");
	neo::cca::ComponentID_shared myself = services->getComponentID();
	neo::cca::ConnectionID_shared goConn = bs->connect(myself, pname, target, port);
	p_go = services->getPort(pname);
	neo::cca::ports::GoPort *go = dynamic_cast< neo::cca::ports::GoPort *>(p_go);
	if (go == 0) {
		throw neo::cca::Exception(neo::cca::BadPortType, "Port is not of expected type GoPort.");
	}
	int result;
	result = go->go();
	if (result != 0) {
		std::cerr << pname <<" returned error code: " << result << std::endl;
	}
	services->releasePort(pname);
	bs->disconnect(goConn, 0.0);
	services->unregisterUsesPort(pname);
	return result;
}

//====================== stuff stolen from ccaffeine ================

#define REFIX(x,y) if (typeName == y)  return neo::cca::x
::neo::cca::Type 
neo::support::helpers::typeFromString( const ::std::string &typeName)
{
	REFIX(NoType,"NoType");
	REFIX(Int,"Int");
	REFIX(Long,"Long");
	REFIX(Float,"Float");
	REFIX(Double,"Double");
	REFIX(Dcomplex,"Dcomplex");
	REFIX(Fcomplex,"Fcomplex");
	REFIX(Bool,"Bool");
	REFIX(String,"String");
	REFIX(IntArray,"IntArray");
	REFIX(LongArray,"LongArray");
	REFIX(FloatArray,"FloatArray");
	REFIX(DoubleArray,"DoubleArray");
	REFIX(DcomplexArray,"DcomplexArray");
	REFIX(FcomplexArray,"FcomplexArray");
	REFIX(BoolArray,"BoolArray");
	REFIX(StringArray,"StringArray");
	return neo::cca::NoType;
}
#undef REFIX

/** key, tm, the destination info. dt, the required type.
 * v, the value to parse into type dt. 
*/
void neo::support::helpers::transformString(const ::std::string & key,
					 ::neo::cca::Type dt,
                     const ::std::string & v, ::neo::cca::TypeMap_shared tm)
{
	// fixme. try parsing string v to type dt and put it in tm
	// at present this has the effect of dropping the string onthe floor
	// if the parse is not supported.
	if (!tm) { return; }
	::std::istringstream ist(v);
#define UNSTRINGVAL(FTYPE,PRIM) \
    case ::neo::cca::FTYPE: \
    { PRIM val = 0; \
      ist >> val; \
      tm->put##FTYPE(key, val); \
    } \
    return 
	switch(dt) 
	{
	case ::neo::cca::String: 
		tm->putString(key, v); 
		return;
	UNSTRINGVAL(Int, int32_t);
	UNSTRINGVAL(Long,  int64_t);
	UNSTRINGVAL(Double, double);
	UNSTRINGVAL(Float, float);
	UNSTRINGVAL(Bool, bool);
	UNSTRINGVAL(Dcomplex, ::std::complex<double> );
	UNSTRINGVAL(Fcomplex, ::std::complex<float> );
	case ::neo::cca::NoType: return ;
	case ::neo::cca::IntArray:
	case ::neo::cca::LongArray:
	case ::neo::cca::FloatArray:
	case ::neo::cca::DoubleArray:
	case ::neo::cca::DcomplexArray:
	case ::neo::cca::FcomplexArray:
	case ::neo::cca::BoolArray: 
	case ::neo::cca::StringArray:
	// fixme whine here.
	default:
		return;
	}
#undef UNSTRINGVAL
}


// YOU DON'T WANT TO TOUCH anything above here. 
// AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
// AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
//
//---------------------------------------------------------- 

int neo::support::helpers::putValueByStrings( ::neo::cca::TypeMap_shared t, const ::std::string & key, const ::std::string &vtype, const ::std::string & val)
{
	if (!t || key.size() == 0) { return -1; }
	::neo::cca::Type kt = typeFromString(vtype); 
	if (kt ==  ::neo::cca::NoType) { return -1; }
	try {
		neo::support::helpers::transformString( key, kt, val, t);
		return 0;
	}
	catch ( std::exception &e )
	{
		return -1;
	}

}

//---------------------------------------------------------- 
const ::std::string 
neo::support::helpers::typeToName( ::neo::cca::Type t) { 
	switch(t) 
	{
		case ::neo::cca::NoType: return "NoType"; 
		case ::neo::cca::Int: return "Int"; 
		case ::neo::cca::Long: return "Long"; 
		case ::neo::cca::Float: return "Float"; 
		case ::neo::cca::Double: return "Double"; 
		case ::neo::cca::Dcomplex: return "Dcomplex"; 
		case ::neo::cca::Fcomplex: return "Fcomplex"; 
		case ::neo::cca::Bool: return "Bool"; 
		case ::neo::cca::String: return "String"; 

		case ::neo::cca::IntArray: return "IntArray"; 
		case ::neo::cca::LongArray: return "LongArray"; 
		case ::neo::cca::FloatArray: return "FloatArray"; 
		case ::neo::cca::DoubleArray: return "DoubleArray"; 
		case ::neo::cca::DcomplexArray: return "DcomplexArray"; 
		case ::neo::cca::FcomplexArray: return "FcomplexArray"; 
		case ::neo::cca::BoolArray: return "BoolArray"; 
		case ::neo::cca::StringArray: return "StringArray"; 
		default: return "unknown";
	}
}
//
//
//---------------------------------------------------------- 
//
::std::string 
neo::support::helpers::stringType( ::neo::cca::TypeMap_shared t, const ::std::string &key)
{
	if (!t) {
		return "";
	}
	enum ::neo::cca::Type kt = t->typeOf(key);
	return neo::support::helpers::typeToName(kt);
}

//---------------------------------------------------------- 
// there's a nice nonportable way to do this with strstream
#define STRINGVAL(BTYPE, FTYPE, DEFVAL, FMT) \
  case ::neo::cca::FTYPE: \
    { \
      BTYPE v = t->get##FTYPE(key, DEFVAL); \
      sprintf(buf, FMT, v); \
      s += buf; \
      return s; \
    } \
  break
#undef STRINGVAL
#define STRINGVAL(FTYPE,DEFVAL) \
  case ::neo::cca::FTYPE: \
    ost << t->get##FTYPE(key, DEFVAL); \
    return ost.str(); \
  break
        
#define ARRAYVAL(FTYPE,VTYPE) \
  case ::neo::cca::FTYPE##Array: \
  { \
    ::std::vector< VTYPE > v; \
    v = t->get##FTYPE##Array( key, v ); \
    for (size_t i = 0; i < v.size(); i++) \
    { \
      ost << v[i] << " "; \
    } \
    return ost.str(); \
  } \
  break
        
::std::string neo::support::helpers::stringValue( ::neo::cca::TypeMap_shared t, const ::std::string &key)
{
	::std::string s;
	if (!t) { return s; }
	::std::ostringstream ost;
	enum ::neo::cca::Type kt = t->typeOf(key);
	// char buf[128];
	switch (kt) {
	case ::neo::cca::NoType: return s;
	STRINGVAL(Int,0);
	STRINGVAL(Long, 0);
	STRINGVAL(Double, 0);
	STRINGVAL(Float, 0);
	STRINGVAL(Bool, false);
	STRINGVAL(String, "");
	STRINGVAL(Dcomplex, ::std::complex<double>(0,0) );
	STRINGVAL(Fcomplex, ::std::complex<float>(0,0) );
	ARRAYVAL(Int, int32_t);
	ARRAYVAL(Long, int64_t);
	ARRAYVAL(Double, double);
	ARRAYVAL(Float, float);
	ARRAYVAL(Bool, bool);
	ARRAYVAL(String, ::std::string );
	ARRAYVAL(Dcomplex, ::std::complex<double> );
	ARRAYVAL(Fcomplex, ::std::complex<float> );
	default:
		return s;
	}
}
#undef STRINGVAL
#undef ARRAYVAL

void 
neo::support::helpers::dumpTypeMapStdout(::neo::cca::TypeMap_shared t)
{
	std::vector< std::string > keys = t->getAllKeys( ::neo::cca::NoType );
	std::sort(keys.begin(), keys.end());
	for (size_t i = 0; i < keys.size(); i++) {
		::std::cout << "["<< keys[i] << "](" <<
		       	neo::support::helpers::stringType(t,keys[i]) << ") " <<
			neo::support::helpers::stringValue(t,keys[i]) << std::endl;
	}
}
