#include <iostream>
#include <cstdio>
#include <neocca.hh>
#include <neoports.hh>
#include "Parameter.hh"

#ifndef lint
namespace {
char id[] =
"$Id: Parameter.cpp,v 1.7 2005/10/10 15:21:43 baallan Exp $";
} 
#endif

using namespace std;

//////////////////////////////////////////
namespace neo {
	namespace test {
class InternalParameterGetListener : public virtual neo::cca::ports::ParameterGetListener
{
private:
	string portName;

public:

	InternalParameterGetListener(const string & pName)
		: portName(pName)
	{
	}

	virtual ~InternalParameterGetListener()
	{
		portName = "dead";
	}

	virtual void updateParameterPort(const string & portName_)
	{
		cout << portName << ".updateParameterPort(" << portName_ << ") called." <<
			endl;
	}
};

//////////////////////////////////////////
class InternalParameterSetListener : public virtual neo::cca::ports::ParameterSetListener
{
private:
	string portName;

public:

	InternalParameterSetListener(const string & pName)
		: portName(pName)
	{
	}

	virtual ~InternalParameterSetListener()
	{
		portName = "dead";
	}

	virtual void updatedParameterValue(string portName_, string fieldName)
	{
		cout << portName << ".updatedParameterValue(" << 
			portName_ << ", " <<
			fieldName << ") called." << endl;

	}
};

}// end namespace test
}// end namespace neo

//////////////////////////////////////////

neo::test::Parameter::Parameter(){ 
	myCore = 0; numtests = 0;
	cout << "Parameter:  CTOR" << endl;
}

neo::test::Parameter::~Parameter(){
	myCore = 0;
	cout << "Parameter:  DTOR" << endl;
}

void
neo::test::Parameter::setServices( neo::cca::Services *cc) 
throw(neo::cca::Exception)
{
	cout << "Parameter:  SET_SERVICES" << endl;
	if (cc == 0)
	{
		cout << "neo::test::Parameter::setServices(0) called." << endl;
		throw neo::cca::Exception("neo::test::Parameter::setServices(0) called.");
	}
	if (myCore != 0) {
		cout << "neo::test::Parameter::setServices() called twice." << endl;
		throw neo::cca::Exception("neo::test::Parameter::setServices() called twice.");
	}

	cout << "Parameter:  startup"  << endl;
	myCore = cc;

  // Contact the PrintfService
	myCore->registerUsesPort( "pSvc", "neo.cca.ports.JPrintfPort" );
	neo::cca::Port * lp = 0;
	try {
		lp = myCore->getPort("pSvc");
	} catch (neo::cca::Exception ex) {
		myCore->addProvidesPort(this, "DEAD=NoJPrintf", "neo::cca::ports::GoPort" );
		cout << "!!! ERROR: No JPrintfService available from framework." << endl;
	}
	jpfp = dynamic_cast< neo::cca::ports::JPrintfPort * >(lp);
	if(jpfp == 0) {
		myCore->addProvidesPort(this, "DEAD=BadJPrintf", "neo::cca::ports::GoPort" );
		cout << "!!! ERROR Uncastable JPrintfService available from framework." << endl;
	}

	myCore->registerUsesPort("ppf", "::neo::cca::ports::ParameterPortFactory");

	myCore->addProvidesPort(this, "go", "neo::cca::ports::GoPort" );
	cout << "Parameter:  startup done"  << endl;
}

int 
neo::test::Parameter::go()
{
  char buf[40];
  numtests++;
  sprintf(buf,"%d",numtests);
  neo::cca::Port * port= 0;


  neo::cca::TypeMap_shared tm = myCore->createTypeMap();

  pplist.push_back(tm);
  string pname = "PP_";
  pname += buf;


  cout << "Parameter:  getting ppf." << endl;
  port = myCore->getPort("ppf");
  if (port == 0) {
	   return -1;
  }
  cout << "Parameter:  got ppf." << endl;
  ppf = dynamic_cast< neo::cca::ports::ParameterPortFactory *>(port);
  ppf->createParameterPort(tm, pname);

  string title = "Test PPF for port ";
  title += pname;
  ppf->setBatchTitle(tm, title);
  ppf->addRequestBoolean(tm,"noName","var to test if default group gets used","anon group",true);
  ppf->setGroupName(tm,"Named Set1");
  ppf->addRequestInt(tm,"iVar","a ranged test integer","int test", 5, 0 ,10);
  ppf->addRequestLong(tm,"jVar","a deranged test long","long test", -50, 0 , -100);
  ppf->setGroupName(tm,"Named Set2");
  ppf->setGroupNameAndTitle(tm,"Named Set2","Pretty title");
  ppf->addRequestDouble(tm,"dVar","a ranged test double","double test", -50, 0 , -100);
  ppf->addRequestFloat(tm,"fVar","a ranged test float","float test", 50, -1000 , 1000);
  ppf->setGroupName(tm,"Named Set3");
  ppf->addRequestString(tm,"sVar","a free test string","string any test", "some value");
  ppf->addRequestString(tm,"sList","a choice test string","string list test", "some value");
  ppf->addRequestStringChoice(tm,"sList","choice 1");
  ppf->addRequestStringChoice(tm,"sList","choice 3");
  ppf->addRequestStringChoice(tm,"sList","choice 2");
  vector< string > moreStrings;
  moreStrings.push_back("fred");
  moreStrings.push_back("barney");
  ppf->addStringChoices(tm,"sList", moreStrings);
  int itest[3] = { 3, 5, 7 };
  ppf->addRequestInt(tm,"icVar","a choiced test integer","int-choice test", 5, 5 ,5);
  ppf->addIntChoices(tm,"icVar", itest, 3);
  
  
  InternalParameterSetListener * setListener = new InternalParameterSetListener(pname);
  InternalParameterGetListener * getListener = new InternalParameterGetListener(pname);

  ppf->registerUpdater(tm, getListener);
  ppf->registerUpdatedListener(tm, setListener);
  setListeners.push_back(setListener);
  getListeners.push_back(getListener);

  ppf->addParameterPort(tm, myCore);
  cout << "Parameter:  configured tm." << endl;
// FIXME enable this
#if 0
  // cycle through a remove just in case we don't get through
  // a destructor protocol due to bugs elsewhere.
  ppf->removeParameterPort(tm, os);
  ppf->addParameterPort(tm, myCore);
#endif
  ppf = 0;

  myCore->releasePort("ppf");
  neo::cca::ports::ParameterPort *pp;
  neo::cca::Port *np = 0;
  np = myCore->getPort(pname);
  if (np == 0) {
    cout << "Parameter: getport(" << pname << ") failed." << endl;
  }
  pp = dynamic_cast< ::neo::cca::ports::ParameterPort *>(np);
  if (pp != 0) {
    if (jpfp) { jpfp->en("Parameter got neo pp"); } else { cout << "NO_JPFP: Parameter got neo pp" << endl; }
    neo::cca::TypeMap_shared ptm2 = pp->readConfigurationMap();
    myCore->releasePort(pname);
    string svar2 = ptm2->getString("sVar","failed svar fetch");
    cout << "Parameter:  sVar = " << svar2 << endl;
  }
  
  return 0;
}

void 
neo::test::Parameter::releaseServices(neo::cca::Services *svc)
throw(neo::cca::Exception)
{
	if (svc != myCore) {
		throw neo::cca::Exception( "neo::test::Parameter::releaseServices svc != myCore");
	}
	if (svc == 0) {
		throw neo::cca::Exception( "neo::test::Parameter::releaseServices called on unstarted/already dead instance.");
	}

	cout << "Parameter:  shutdown"  << endl;
	neo::cca::Port * port;
	port = myCore->getPort("ppf");
	if (port != 0) {
		cout << "Parameter: ppf fetched in setservices. " << endl;
		ppf = dynamic_cast< ::neo::cca::ports::ParameterPortFactory *>(port);
		if (!ppf == 0) {
			for (size_t i = 0; i < pplist.size(); i++)
			{
				neo::cca::TypeMap_shared tm = pplist[i];
				ppf->removeParameterPort(tm, myCore);
				InternalParameterGetListener * gl = getListeners[i];
				InternalParameterSetListener * sl = setListeners[i];
				if (sl != 0) { delete sl; }
				if (gl != 0) { delete gl; }
			}
			setListeners.clear();
			getListeners.clear();
		}
		myCore->releasePort("ppf");
	} else {
		cout << "Parameter: ppf fetch failed in releaseServices. " << endl;
	}
	myCore->unregisterUsesPort("ppf");
	myCore->removeProvidesPort("go");
	myCore = 0;
	return;
}

extern "C" {

neo::cca::Component *create_neo_test_Parameter() {
	neo::cca::Component *comp_ptr;
	neo::test::Parameter *component;
	component = new neo::test::Parameter();
	comp_ptr = dynamic_cast<neo::cca::Component *>(component);
	return comp_ptr;
}


}

